import datetime

def dateGenerator(start_date, end_date, days_frequency):
    """
    Creating a range of dates dependent upon the frequency as set by the user.

    Input Parameters:
        1. start_date     => The starting date for frequency calculations
        2. end_date       => The ending date, the antichrist
        3. days_frequency => The frequency of days when Boromir does not drink
    """

    start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d")
    end_date   = datetime.datetime.strptime(end_date, "%Y-%m-%d")

    current_date = start_date
    while current_date < end_date:
        next_date = current_date + datetime.timedelta(days=days_frequency)
        if next_date > end_date: next_date = end_date

        yield [datetime.datetime.strftime(current_date, "%Y-%m-%d"), datetime.datetime.strftime(next_date, "%Y-%m-%d")]
        current_date = next_date
        pass
    pass


"""

NOTE: This code only runs if kawaplatformlibrary is already installed in the virtual environment.

"""

"""
Example for downloading Sentinel 2 tiles for a specific AOI. All parameters except the TCI flag require values set by the user.

Input parameters:
    1. geojson                  => GeoJSON for the Area of Interest[AOI]. Optional with a default value
    2. start_date               => The initial date from when data should be requested from STAC API. Format maintained is YYYY-MM-DD. Optional with a default value.
    3. end_date                 => The final date before which data should be requested from STAC API. Format maintained is YYYY-MM-DD. Optional with a default value.
    4. bands                    => The bands to be requested from the STAC API for the Sentinel 2 tiles. Comma separated list. Optional with default being ["B02","B03","B04"]
    5. ground_sampling_distance => The ground sampling distance to which all bands should be resampled to. Optional with a default value of 10m.
    6. destination_file         => The file location where the obtained TIF file should be stored.

Output parameters:
    Stores the data obtained in a file specified by the user.

Example UseCase : 
python test_download_tiles.py --geojson ../../test_aoi.geojson --start_date 2020-02-01 --end_date 2020-02-27 --bands B02,B03,B04 --ground_sampling_distance 10 --destination_file ../../Test_Examples_File_Path.tif --num_threads 6
"""

# importing libraries
import numpy as np
import json, argparse, rasterio, os

import sys
sys.path.insert(1, "./../") # Changing the system path to root directory of the library. This is needed to actually call the Kawa Platform library

from kawaplatformlibrary.data import sentinel2dataingestion
from kawaplatformlibrary.data.preprocessing import splitgeojson
from kawaplatformlibrary.postprocessing.mosiac import mergeBands, createMosiacTiles
from kawaplatformlibrary.postprocessing.createprofile import createNewProfile
from kawaplatformlibrary.postprocessing.creatememoryraster import createMemRaster

from rasterio import Affine

# Optional arguments for the user
ap = argparse.ArgumentParser()
ap.add_argument("-g", "--geojson", required=True, 
    help="Path to GEOJSON file containing the AOI.")
ap.add_argument("-s", "--start_date", required=True, 
    help="Start date in YYYY-MM-DD format")
ap.add_argument("-e", "--end_date", required=True, 
    help="End date in YYYY-MM-DD format")
ap.add_argument("-b", "--bands", required=True,
    help="Comma separated list of bands")
ap.add_argument("-t", "--tci", nargs="?", default=False, type=bool,
    help="If True Colour Image is required.")
ap.add_argument("-d","--ground_sampling_distance", required=True, type=int,
    help="Ground Sampling Distance required for each band")
# This is not optional. User should input this
ap.add_argument("-f", "--destination_folder", required=True,
    help="Path to file for storage. Use the full OS path and not relative path")
ap.add_argument("-n", "--num_threads", required=True, type=int,
    help="Number of parallel process to run.")
ap.add_argument("-i", "--days_interval", default=None,
    help="Interval for obtaining data.")
args = vars(ap.parse_args())

if args["days_interval"]:
    args["days_interval"] = int(args["days_interval"])

user_bands = [band.strip() for band in args["bands"].split(',')]

# Reading the GeoJSON of the AOI and extracting the coordinates json. {"type": "Polygon", "coordinates":[[[...], [...], ...]]]}
with open(args["geojson"], "r") as in_file:
    geojson_contents = json.load(in_file)
    geojson_coordinates = geojson_contents["features"][0]["geometry"]
    in_file.close()
    pass

# Obtaining image url from STAC API. dataCheck = [0/1, "[ERROR]/["INFO"]", {}/{<data>}]
print("[INFO] Obtianing data for the AOI")
sentinel2_data_class_obtain_data = sentinel2dataingestion.ObtainData(bands=user_bands, TCI=args["tci"], cloud_cover=10)
print("[INFO] Finished finding data for the AOI.")

def downloadData(rasters_href):
    """
    Downloading and storing the data in the destination folder.
    """
    print("[INFO] Downloading data for {}".format(rasters_href["img_id"]))

    destination_file = rasters_href["img_id"] + ".tif"
    destination_file = os.path.join(args["destination_folder"], destination_file)

    bands_rasters_list = []
    bands_data = rasters_href["band_data"]

    # Reading the bands into memory as numpy arrays
    for j, band in enumerate(bands_data):

        band_href = bands_data[band]["href"]
        band_src = rasterio.open(band_href)
        bands_rasters_list.append(band_src)

        # Creating the destination profile for each Sentinel 2 tile. Will be used for storing the current tile.
        if j == 0:
            num_bands = len(user_bands)
            dst_profile = createNewProfile(band_src.transform, band_src.profile, ground_sampling_distance=10, num_bands=num_bands)
            pass
        pass

    # Resampling and merging the bands
    merged_resample_bands = mergeBands(bands_rasters_list, args["ground_sampling_distance"])
    print("[INFO] Finished merging bands for img : {}".format(rasters_href["img_id"]))

    bands_raster = createMemRaster(dst_profile, merged_resample_bands)
    
    # Storing the Sentinel 2 tile in the destination folder with tile name corresponding to the tile ID
    with rasterio.open(destination_file, "w", **dst_profile) as out_file:
        for i in range(merged_resample_bands.shape[0]):
            out_file.write(merged_resample_bands[i].astype(rasterio.float32), i+1)
            pass
        pass

    print("[INFO] Finished downloading data for {}".format(rasters_href["img_id"]))
    return bands_raster
    pass

list_rasters = []

if args["days_interval"]:
    
    sentinel2_data_aoi_date = sentinel2_data_class_obtain_data.getData(aoi_geojson=geojson_coordinates, start_date=args["start_date"], end_date=args["end_date"], days_frequency=args["days_interval"], num_threads=args["num_threads"])
    
    for sentinel2_data_aoi in sentinel2_data_aoi_date:
        for sentinel2_data in sentinel2_data_aoi:
            data = list(sentinel2_data.items())[0][1]
            data_sensing_date = list(sentinel2_data.items())[0][0]
            data_check = data[0]

            if data_check == 1:
                list_rasters.append(downloadData(data[2]))
                pass
            else:
                print("[INFO] No images found for date {}. Error Message : {}".format(data_sensing_date, data[0]))
    pass
else:

    sentinel2_data_aoi = sentinel2_data_class_obtain_data.getData(aoi_geojson=geojson_coordinates, start_date=args["start_date"], end_date=args["end_date"], num_threads=args["num_threads"])

    for sentinel2_data in sentinel2_data_aoi:
        data = list(sentinel2_data.items())[0][1]
        data_sensing_date = list(sentinel2_data.items())[0][0]
        data_check = data[0]

        if data_check == 1:
            list_rasters.append(downloadData(data[2]))
            pass
        else:

            print("[INFO] No images found. Error Message : {}".format(data[1]))
        pass
    pass

print("[INFO] Merging all rasters.")
merged_raster = createMosiacTiles(list_rasters)
print(merged_raster.profile)
print("[END] Finished merging rasters")


"""
Calculation of Normalised Difference Vegetation Index[NDVI] using the Near-Infrared and Infrared bands

Formula : (NIR - RED)/(NIR + RED)
Wavelenghts : 640:760nm, 780:1400nm
"""

import numpy as np
import pandas as pd
import rasterio, argparse, os

from kawaplatformlibrary.postprocessing.mosiac import reSampleBand

def store():
    """
    This will be done when we have decided on the data storage strategy.
    """
    pass

def calculate(band_red, band_nir, ground_sampling_distance):
    """
    Function for calculating the NDVI
    
    Input Parameters:
        1. band_nir => NIR band of the satellite image
        2. band_red => Red band of the satellite image

    Output Parameters:
        1. ndvi => NDVI for the entire satellite image
    """

    if band_nir.transform[0] != ground_sampling_distance:
        band_nir_array = reSampleBand(band_nir, ground_sampling_distance)
    else:
        band_nir_array = np.array(band_nir.read(1), dtype=np.float32)

    if band_red.transform[0] != ground_sampling_distance:
        band_red_array = reSampleBand(band_red, ground_sampling_distance)
    else:
        band_red_array = np.array(band_red.read(1), dtype=np.float32)

    ndvi = (band_nir_array - band_red_array)/(band_nir_array + band_red_array)

    return np.array(ndvi, dtype=np.float32)
    pass



"""
Calculation of Canopy Chlorophyll Content Index using the Near-Infrared, Red and Red-Edge bands

Formula: ((NIR - RED_EDGE) / (NIR + RED_EDGE)) / ((NIR - RED) / (NIR - RED))

Wavelengths: 690nm :730nm , 780nm :1400nm
"""

import numpy as np
import pandas as pd
import rasterio, os

from kawaplatformlibrary.postprocessing.mosiac import reSampleBand

def store():
    """
    This will be done when we have decided on the data storage strategy.
    """
    pass

def calculate(band_red, band_red_edge, band_nir, ground_sampling_distance):
    """
    Function for calculating the CCCI
    
    Input Parameters:
        1. band_nir => NIR band of the satellite image
        2. band_red => Red band of the satellite image
        3. band_red_edge => Red-edge band of the satellite image
    Output Parameters:
        1. ccci => CCCI for the entire satellite image
    """

    if band_nir.transform[0] != ground_sampling_distance:
        band_nir_array = reSampleBand(band_nir, ground_sampling_distance)
    else:
        band_nir_array = np.array(band_nir.read(1), dtype=np.float32)

    if band_red.transform[0] != ground_sampling_distance:
        band_red_array = reSampleBand(band_red, ground_sampling_distance)
    else:
        band_red_array = np.array(band_red.read(1), dtype=np.float32)

    if band_red_edge.transform[0] != ground_sampling_distance:
        band_red_edge_array = reSampleBand(band_red_edge, ground_sampling_distance)
    else:
        band_red_edge_array = np.array(band_red_edge.read(1), dtype=np.float32)

    ccci = ((band_nir_array-band_red_edge_array)/(band_nir_array+band_red_edge_array))/((band_nir_array-band_red_array)/(band_nir_array-band_red_array))

    return np.array(ccci, dtype=np.float32)
    pass



import numpy as np

from pyproj import Transformer
from shapely.geometry import Polygon, Point
from rasterio.mask import mask

from .createprofile import createNewProfile
from .creatememoryraster import createMemRaster

def reprojectPolygon(gj, outproj, inproj):
    '''
    Reprojecting a GeoJson bounding box from one projection to another

    Input Parameters:
        1. gj      => GeoJson containing the coordinates to be reprojected
        2. outproj => Projection to be converted into
        3. inproj  => Current projection of the coordinates

    Output
        Reprojected GeoJson
    '''

    min_lat = gj['coordinates'][0][2][1]
    max_lat = gj['coordinates'][0][0][1]
    # print("Min Lat: {}, Max Lat: {}".format(min_lat, max_lat))

    max_lon = gj['coordinates'][0][2][0]
    min_lon = gj['coordinates'][0][0][0]
    # print("Min Lon: {}, Max Lon: {}".format(min_lon, max_lon))

    lat1_list = [min_lat, min_lat, max_lat, max_lat, min_lat]
    lon1_list = [min_lon, max_lon, max_lon, min_lon, min_lon]
    
    transformer = Transformer.from_crs(inproj, outproj)
    
    lon_list, lat_list = transformer.transform(lat1_list,lon1_list)
    
    polygon_reproj = Polygon(zip(lon_list, lat_list))

    polyg = polygon_reproj.exterior.coords

    coords = []
    for x,y in polyg:
        coords.append([x,y])
        pass

    gj_reproj = {"type": "Polygon", "coordinates" : [coords]}
    return gj_reproj
    pass

def extractArea(area_coordinates, raster):
    """
    Function for extracting an area from a raster and passing back a memory raster.

    Input Parameters:
        1. area_coordinates => Dictionary containing the coordinates and type of coordinates. Can also be a shapely polygon. CRS for the coordinates should be the same as the raster.
        2. raster           => Either memory raster or an opened TIF file

    Output:
        1. 
    """

    area_img, area_trans = mask(raster, [area_coordinates], crop=True)

    area_profile = createNewProfile(area_trans, raster.profile, img_height=area_img.shape[1], img_width=area_img.shape[2], num_bands=area_img.shape[0])

    return area_img, area_profile
    pass


"""
Creation of a data ingestion pipeline for Sentinel 2 Level 2 A data from AWS S3 bucket using STAC api.

Calls CustomExceptions to raise errors with user input data.
"""

import datetime
import os
import requests

from shapely.geometry import shape
from rasterio import Affine

from kawaplatformlibrary.data.preprocessing import splitgeojson
from kawaplatformlibrary.data.preprocessing import splitdates
from joblib import Parallel, delayed

class ObtainData():
    """
    Class for obtaining data from Sentinel 2 STAC API
    """

    def __init__(self, bands=["B01", "B02", "B03", "B04", "B05", "B06", "B07", "B08", "B09", "B11", "B12", "B8A"], cloud_cover=5, area_coverage=90, intersection_area=20, TCI=False):
        """
        Input Parameters:
            1. bands             => List of bands from the Sentinel 2 iamge. B10 is not included in L2A
            2. cloud_cover       => Maximum cloud cover for the Sentinel 2 image
            3. area_coverage     => Minimum amount of data to be present in the image
            4. intersection_area => Minumum intersection between Bounding Box of Sentinel 2 tile and query geojson
            5. TCI               => If the True Colour Image is required or not.
        """

        self.bands             = bands
        self.cloud_cover       = cloud_cover
        self.area_coverage     = area_coverage
        self.intersection_area = intersection_area
        self.TCI               = TCI
        pass

    def getData(self, aoi_geojson, start_date, end_date, days_frequency=None, num_threads=None):
        """
        Getting the bands and their data from Sentinel 2 STAC API using multiprocessing.

        Input Parameters:
            1. aoi_geojson    => The AOI for which Sentinel 2 data is to be obtained
            2. start_date     => Starting date for inference period
            3. end_date       => Ending date for inference period
            4. days_frequenct => Intervals or frequency of Boromir drinking water instead of beer.
            4. num_threads    => Number of processes to run paralelly to obtain the data from S2 STAC API.
            
        Output:
            1. Returns a dictionary with key being the date and value being the bands data, code, and message.
            [NOTE] If days_frquency is provided, this function will yield values and not return values.
        References:
            1. ObtianData.getBandData()
        """

        # If num_threads is None or is more than the numebr of threads present on the CPU, the code uses the maximum number of threads available on the system
        if num_threads == None or num_threads > os.cpu_count():
            num_threads = os.cpu_count()
            pass

        # If days_frequency is less than 5, setting the frequency to 5 which is the temporal resolution of the satellite.
        if days_frequency != None:
            if days_frequency < 5: days_frequency = 5

        # Splitting the larger GeoJSON into smaller GeoJSON.
        split_geojson_coords_list = splitgeojson.split("SENTINEL2", aoi_geojson)
        
        if days_frequency:
            aoi_data_list = []
            for start_date_freq, end_date_freq in splitdates.dateGenerator(start_date, end_date, days_frequency):
                aoi_data_list.append(Parallel(n_jobs=num_threads)(delayed(self.getBandData)(split_geojson_polygon, start_date_freq,
                                    end_date_freq, i+1) for i,split_geojson_polygon in enumerate(split_geojson_coords_list)))
                pass
            return aoi_data_list
        else:
            return Parallel(n_jobs=num_threads)(delayed(self.getBandData)(split_geojson_polygon, start_date, end_date, i+1) for i, split_geojson_polygon in enumerate(split_geojson_coords_list))

        pass

    def getBandData(self, geojson, start_date, end_date, tile_number = 1):
        """
        Function for obtaining metadata from Sentinel 2 AWS bucket for the specific AOI 
        
        The output is a list of 3 elements. The first element is a number specifying whether the call worked or not. The second element is a text message explaining the first element. The third element is the data to be used by the user.
        
        Error messages will be numbered evens while info messages will be numbered in odd.

        Input Parameters:
            1. geojson     => The AOI for which Sentinel 2 data should be found
            2. tile_number => The tile number from split_geojson. Defaults to one.

        Output:
             Info message stating that data has been found or there is an error with the third element either being the tile_number[ERROR] or a dictionary containing the following[DATA FOUND]:
                2.1. img_id        => Sentinel 2 image ID
                2.2. band_date     => Dictionary if the user specified bands, each band is a dictionary containing the URL and Affine transform for that band.
                2.3. "init : EPSG" => The CRS for that AOI, corresponds to different UTM zones.
                2.4. bbox          => Bounding box of the Sentinel 2 tile. May not correspond to the actual tile.
                2.5. sensing_date  => The date when the data was obtained in miliseconds since epoch
                2.6. TCI           => If TCI is specified by the user, it will be added otherwise it is not.
        
        Called:
            1. ObtainData.getBandData()

        ERROR Codes:
            1. 0 =>  Band specified by user is not present in Sentinel 2 STAC API.
            2. 2 =>  GeoJSON type is not a Polygon
            3. 4 =>  URL could not be accessed due to errorneous user input
            4. 6 =>  No images could be found
        
        INFO Codes
            1. 1 => Images found and links sent as a dictionary
        """

        # Default list of bands for checking user inputted bands
        default_bands = ["B01", "B02", "B03", "B04", "B05", "B06", "B07", "B08", "B09", "B11", "B12", "B8A"] 

        # Checking if bands given are correct or not. Error passed when user inputed band is incorrect as per default_bands list
        for user_band in self.bands:
            if user_band not in default_bands:
                return {start_date : [0, "[ERROR] Band is not present in Sentinel 2 Level 2A, User band: {}".format(user_band), tile_number]}
                pass
            pass

        # Reading the contents of the GeoJSON and extracting the coordinates
        # Removed reading file as it's redundant
        
        bounding_box_type = geojson["type"]
        
        # Checking the geometry type of the bounding box. Geometry type should be Polygon
        if bounding_box_type != "Polygon":
            return {start_date : [2, "[ERROR] Required GeoJSON type to be a Polygon.", tile_number]}
            pass

        bounding_box = [str(coord) for coord in shape(geojson).bounds]
        bounding_box = ",".join(bounding_box)

        shape_aoi = shape(geojson)

        url = f"https://earth-search.aws.element84.com/v0/collections/sentinel-s2-l2a-cogs/items?limit=100&bbox=[{bounding_box[:-1]}]&datetime={start_date}T00:00:00Z/{end_date}T23:59:59Z"
        
        try:
            search_response = requests.get(url).json()
        except:
            return {start_date : [4, "[ERROR] URL could not be accessed", tile_number]}
            pass

        items = search_response['features']

        img_list = []

        # Taking starting date as the minimum date
        min_datetime = datetime.datetime.strptime(start_date, "%Y-%m-%d")
        min_datetime = min_datetime.timestamp()

        for item in items:

            shape_item = shape(item["geometry"])

            # If the intersection of the Sentinel 2 tile with the current AOI is less than 30% then skip the tile.
            if (shape_aoi.intersection(shape_item).area/shape_aoi.area)*100 < self.intersection_area:
                continue
                pass

            if item['properties']["eo:cloud_cover"] < self.cloud_cover:
                
                # Two if statements to accept the different data_coverage keys as per STAC API.

                if "data_coverage" in item['properties'] and item['properties']['data_coverage'] > self.area_coverage:
                    
                    # Storing the band href link and transform in a dictionary
                    band_data = {band:{'href':item["assets"][band]["href"], 'affine_transform':Affine(*item["assets"][band]["proj:transform"][:6])} for band in item["assets"] if band in self.bands}
                    
                    # Storing the EPSG in crs format as well as a tuple of the bounding box
                    epsg_data = {"init" : "EPSG:" + str(item["properties"]["proj:epsg"])}
                    bbox = tuple(item["bbox"])
                    
                    date_sensing = item["properties"]["datetime"].split("T")[0]
                    date_sensing = datetime.datetime.strptime(date_sensing, "%Y-%m-%d")
                    date_sensing = date_sensing.timestamp() # converting the datetime string into seconds since epoch

                    # Storing it as per the unique Sentinel 2 ID. If TCI is asked, TCI is added to the dictionary separately.
                    if self.TCI:
                        img_list.append({"img_id" : item['id'], "band_data" : band_data, "init" : "EPSG:" + str(item["properties"]["proj:epsg"]), "bbox" : bbox, "sensing_date" : date_sensing, "TCI" : item["assets"]["visual"]["href"]})
                    else:
                        img_list.append({"img_id" : item['id'], "band_data" : band_data, "init" : "EPSG:" + str(item["properties"]["proj:epsg"]), "bbox" : bbox, "sensing_date" : date_sensing})
                    pass

                if "sentinel:data_coverage" in item['properties'] and item['properties']['sentinel:data_coverage'] > self.area_coverage:    

                    # Storing the band href link and transform in a dictionary
                    band_data = {band:{'href':item["assets"][band]["href"], 'affine_transform':Affine(*item["assets"][band]["proj:transform"][:6])} for band in item["assets"] if band in self.bands}
                    
                    # Storing the EPSG in crs format as well as a tuple of the bounding box
                    epsg_data = {"init" : "EPSG:" + str(item["properties"]["proj:epsg"])}
                    bbox = tuple(item["bbox"])

                    date_sensing = item["properties"]["datetime"].split("T")[0]
                    date_sensing = datetime.datetime.strptime(date_sensing, "%Y-%m-%d")
                    date_sensing = date_sensing.timestamp() # Converting the datetime string into seconds since epoch
                    
                    # Storing it as per the unique Sentinel 2 ID. If TCI is asked, TCI is added to the dictionary separately.
                    if self.TCI:
                        img_list.append({"img_id" : item['id'],"band_data" : band_data, "init" : "EPSG:" + str(item["properties"]["proj:epsg"]), "bbox" : bbox, "sensing_date" : date_sensing, "TCI" : item["assets"]["visual"]["href"]})
                    else:
                        img_list.append({"img_id" : item['id'],"band_data" : band_data, "init" : "EPSG:" + str(item["properties"]["proj:epsg"]), "bbox" : bbox, "sensing_date" : date_sensing})
                    pass

                pass
            pass

        # Checking if images are found
        if not img_list:
            return {start_date : [6, "[ERROR] No images could be found", tile_number]}
            pass

        choice = 0 # Variable for choosing the most recent image
        chosen_date = start_date
        # Choosing the most recent image
        for c, img_data in enumerate(img_list):
            sensing_date = img_data["sensing_date"]
            if sensing_date > min_datetime:
                min_datetime = sensing_date
                chosen_date = datetime.datetime.utcfromtimestamp(int(min_datetime)).strftime("%Y-%m-%d")
                choice = c

        return {chosen_date : [1, "[INFO] Found an image", img_list[choice]]}
        pass
    pass


"""
Calculation of Normalised Difference Water Index (Surface Water) using Near Infrared and Green bands

Formula : (NIR - GREEN) / (NIR + GREEN)

Wavelengths : 490nm: 570nm, 780nm :1400nm
"""

import numpy as np
import pandas as pd
import rasterio, os

from kawaplatformlibrary.postprocessing.mosiac import reSampleBand

def store():
    """
    This will be done when we have decided on the data storage strategy.
    """
    pass

def calculate(band_green, band_nir, ground_sampling_distance):
    """
    Function for calculating the NDWI (Surface water)
    
    Input Parameters:
        1. band_nir => NIR band of the satellite image
        2. band_green => Green band of the satellite image

    Output Parameters:
        1. ndwi_sw => NDWI (Surface water) for the entire satellite image
    """

    if band_nir.transform[0] != ground_sampling_distance:
        band_nir_array = reSampleBand(band_nir, ground_sampling_distance)
    else:
        band_nir_array = np.array(band_nir.read(1), dtype=np.float32)

    if band_green.transform[0] != ground_sampling_distance:
        band_green_array = reSampleBand(band_green, ground_sampling_distance)
    else:
        band_green_array = np.array(band_green.read(1), dtype=np.float32)

    ndwi_sw = (band_green_array - band_nir_array)/(band_green_array + band_nir_array)

    return np.array(ndwi_sw, dtype=np.float32)
    pass



"""
Creating a memory raster for merging. A separate file and function as it does not come under the umbrella of merging or creating a new raster profile.
"""

import numpy
import rasterio

from rasterio.io import MemoryFile

def createMemRaster(src_profile, numpy_array):
    """
    Function for creating a new profile.

    Input parameters : 
        1. src_profile => Profile to which the raster should be written
        2. numpy_array => Numpy array that has to be written to a raster file.

    Output :
        1. dataset => raster created using the src_profile and numpy_array
    """
    with MemoryFile() as memfile:
        dataset = memfile.open(**src_profile)
        dataset.write(numpy_array)
        del numpy_array
        pass

    return dataset
    pass


"""
Function for Resampling bands, merging bands into one numpy array with a common ground sampling distance and merging a list of rasters to create one merged raster.
"""

import numpy as np
import rasterio 

from rasterio.warp import reproject, Resampling
from rasterio import transform
from rasterio.merge import merge
from rasterio import Affine

from .creatememoryraster import createMemRaster

def reSampleBand(band_raster, ground_sampling_distance):
    """
    Function for resampling the bands to a common resolution
    """

    scale = band_raster.transform[0] // ground_sampling_distance

    org_transform = band_raster.transform
    resample_transform = Affine(org_transform.a / scale, org_transform.b, org_transform.c, org_transform.d, org_transform.e / scale, org_transform.f)

    resample_height = int(band_raster.height * scale)
    resample_width = int(band_raster.width * scale)

    data = band_raster.read( # Note changed order of indexes, arrays are band, row, col order not row, col, band
            out_shape=(resample_height, resample_width),
            resampling=Resampling.cubic)

    return np.array(data[0], dtype=np.float32)
    pass

def mergeBands(list_bands_rasters, ground_sampling_distance):
    """
    Function for merging the bands into one numpy array
    """

    for i in range(len(list_bands_rasters)):
        if int(list_bands_rasters[i].transform[0]) != ground_sampling_distance:
            
            list_bands_rasters[i] = reSampleBand(list_bands_rasters[i], ground_sampling_distance)
        else:
            list_bands_rasters[i] = np.array(list_bands_rasters[i].read(1), dtype=np.float32)
        pass

    merged_bands = np.stack(list_bands_rasters)
    return merged_bands
    pass

def createMosiacTiles(list_rasters):
    """
    Function for creating the mosiac of all tiles. It currently takes all rasters and gives the transformation matrix of the mergerd rasters and an array of all the merged rasters.

    Input Parameters:
        1. list_rasters => list of rasters which have to be merged.

    Ouptut:
        1. Merged raster array and it's corresponding profile
    """

    merged_rasters_array, merged_rasters_trans = merge(list_rasters)

    merged_rasters_array_height, merged_rasters_array_width = merged_rasters_array.shape[1], merged_rasters_array.shape[2]

    dst_profile = list_rasters[0].profile

    merged_rasters_profile = dst_profile
    merged_rasters_profile.update(transform=merged_rasters_trans, driver="GTIFF", height=merged_rasters_array_height,
        width=merged_rasters_array_width)

    return merged_rasters_array, merged_rasters_profile
    pass


"""
Calculation of Disease Water Stres Index[DSWI] using 547nm, 682nm, 802nm, 1657nm

Formula : (802 + 547)/(1657 - 682)
Wavelengths : 547nm, 682nm, 802nm, 1657nm
"""

import numpy as np
import pandas as pd
import rasterio, os

from kawaplatformlibrary.postprocessing.mosiac import reSampleBand

def store():
    """
    This will be done when we have decided on the data storage strategy.
    """
    pass

def calculate(band_547, band_682, band_802, band_1657, ground_sampling_distance):
    """
    Function for calculating the DWSI
    
    Input Parameters:
        1. band_547 => 550nm band of the satellite image
        2. band_682 => 670nm band of the satellite image
        3. band_802 => 800nm band of the satellite image
        4. band_1657 => 1657nm band of the satellite image

    Output Parameters:
        1. dwsi => DWSI for the entire satellite image
    """

    if band_547.transform[0] != ground_sampling_distance:
        band_547_array = reSampleBand(band_547, ground_sampling_distance)
    else:
        band_547_array = np.array(band_547.read(1), dtype=np.float32)    

    if band_682.transform[0] != ground_sampling_distance:
        band_682_array = reSampleBand(band_682, ground_sampling_distance)
    else:
        band_682_array = np.array(band_682.read(1), dtype=np.float32)    

    if band_802.transform[0] != ground_sampling_distance:
        band_802_array = reSampleBand(band_802, ground_sampling_distance)
    else:
        band_802_array = np.array(band_802.read(1), dtype=np.float32)    

    if band_1657.transform[0] != ground_sampling_distance:
        band_1657_array = reSampleBand(band_1657, ground_sampling_distance)
    else:
        band_1657_array = np.array(band_1657.read(1), dtype=np.float32)

    dwsi = (band_802_array + band_547_array)/(band_1657_array - band_682_array)

    return np.array(dwsi, dtype=np.float32)
    pass
